# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Any, Literal

    from sqlalchemy.orm import Session

    from rucio.common.types import InternalScope


class DidMetaPlugin(metaclass=ABCMeta):
    """
    Interface for plugins managing metadata of DIDs
    """

    def __init__(self):
        """
        Initializes the plugin
        """
        self._plugin_name = None

    @property
    def name(self) -> str:
        """
        The getter method for the plugin's name.

        :returns: The standardized (casefolded) name of this plugin.
        :raises AttributeError: If '_plugin_name' is not defined in the subclass.
        """
        if self._plugin_name:
            return self._plugin_name.casefold()
        raise AttributeError("Subclasses of DidMetaPlugin must define the '_plugin_name' attribute.")

    def is_named(self, plugin_name: str) -> bool:
        """Return whether the plugin matches the provided name using case-insensitive comparison."""
        return self.name == plugin_name.casefold()

    @abstractmethod
    def get_metadata(
        self,
        scope: "InternalScope",
        name: str,
        *,
        session: "Session | None" = None
    ) -> "Any":
        """
        Get data identifier metadata

        :param scope: The scope of the DID.
        :param name: The data identifier name.
        :param session: The database session in use.
        """
        pass

    @abstractmethod
    def set_metadata(
        self,
        scope: "InternalScope",
        name: str,
        key: str,
        value: str,
        recursive: bool = False,
        *,
        session: "Session | None" = None
    ) -> None:
        """
        Add metadata to data identifier.

        :param scope: The scope of the DID.
        :param name: The data identifier name.
        :param key: the key.
        :param value: the value.
        :param recursive: Instruction to propagate the metadata change recursively to content (False by default).
        :param session: The database session in use.
        """
        pass

    @abstractmethod
    def set_metadata_bulk(
        self,
        scope: "InternalScope",
        name: str,
        meta: dict[str, "Any"],
        recursive: bool = False,
        *,
        session: "Session | None" = None
    ) -> None:
        """
        Add metadata to data identifier in bulk.

        :param scope: The scope of the DID.
        :param name: The data identifier name.
        :param meta: all key-values to set.
        :type meta: dict
        :param recursive: Instruction to propagate the metadata change recursively to content (False by default).
        :param session: The database session in use.
        """
        for key, value in meta.items():
            self.set_metadata(scope, name, key, value, recursive=recursive, session=session)

    @abstractmethod
    def delete_metadata(
        self,
        scope: "InternalScope",
        name: str,
        key: str,
        *,
        session: "Session | None" = None
    ) -> None:
        """
        Deletes the metadata stored for the given key.

        :param scope: The scope of the DID.
        :param name: The name of the DID.
        :param key: The key to be deleted.
        :param session: The database session in use.
        """
        pass

    @abstractmethod
    def list_dids(
        self,
        scope: "InternalScope",
        filters: dict[str, "Any"],
        did_type: "Literal['all', 'collection', 'dataset', 'container', 'file']" = 'collection',
        ignore_case: bool = False,
        limit: "int | None" = None,
        offset: "int | None" = None,
        long: bool = False,
        recursive: bool = False,
        *,
        session: "Session | None" = None
    ) -> "Iterator[str | dict[str, Any]]":
        """
        Search data identifiers

        :param scope: The scope of the DID.
        :param filters: dictionary of attributes by which the results should be filtered.
        :param did_type: the type of the DID: all(container, dataset, file), collection(dataset or container), dataset, container, file.
        :param ignore_case: ignore case distinctions.
        :param limit: limit number.
        :param offset: offset number.
        :param long: Long format option to display more information for each DID.
        :param session: The database session in use.
        :param recursive: Recursively list DIDs content.
        """
        pass

    @abstractmethod
    def manages_key(
        self,
        key: str,
        *,
        session: "Session | None" = None
    ) -> bool:
        """
        Returns whether key is managed by this plugin or not.

        :param key: Key of the metadata.
        :param session: The database session in use.
        :returns: (Boolean)
        """
        pass



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

import pytest
from werkzeug import exceptions

from rucio.web.rest.flaskapi.v1.common import param_get_bool


@pytest.mark.parametrize(
    'params,kwargs,raise_log,raise_error',
    [
        ({"object": False}, {}, False, False),
        ({"object": True}, {}, False, False),
        ({"object": "true"}, {}, False, False),
        ({"object": "false"}, {}, False, False),
        ({"object": '0'}, {}, True, False),
        ({"object": '1'}, {}, True, False),
        ({"object": 1}, {}, True, False),
        ({"object": 0}, {}, True, False),
        ({"object": 'Incorrect Type'}, {}, False, True),
        ({}, {}, False, True),
        ({}, {"default": True}, False, False),
        ({"object": None}, {"default": False}, False, True),
        ({}, {"default": "Incorrect Type"}, False, True),
        ({"object": "trUE"}, {}, False, False),
        ({"object": "FALSE"}, {}, False, False),
    ],
    ids=[
        "Basic Bool - False",
        "Basic bool - True",
        "String Bool - True",
        "String Bool - False",
        "Int-like str bool - False",
        "Int-like str bool - True",
        "Int bool - False",
        "Int bool - True",
        "Type Error",
        "No object, no default",
        "No object, yes default",
        "None object, yes default",
        "No object, invalid default type",
        "Str bool, variant case - True",
        "Str bool, variant case - False"
    ]
)
def test_param_get_bool(params, kwargs, raise_log, raise_error, caplog):
    with caplog.at_level(logging.WARNING):
        if raise_error:
            with pytest.raises(exceptions.HTTPException):
                param_get_bool(params, "object", **kwargs)

        else:
            out = param_get_bool(params, "object", **kwargs)
            assert isinstance(out, bool)

    if raise_log:
        assert "Booleans should only accept true/false. Please change 0/1 to true/false." in caplog.text



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Literal, Optional

import click
from rich.text import Text
from tabulate import tabulate

from rucio.client.richclient import CLITheme, generate_table, print_output
from rucio.common.exception import InputValidationError
from rucio.common.utils import get_bytes_value_from_string, sizefmt


@click.group()
def account():
    """Methods to add or change accounts for users, groups, and services. Used to assign privileges"""


@account.command("add")
@click.argument("account-name")
@click.argument("account-type", type=click.Choice(["USER", "GROUP", "SERVICE"]))
@click.option("--email", type=str, help="Email address associated with the account")
@click.pass_context
def add_(ctx: click.Context, account_type: Literal['USER', 'GROUP', 'SERVICE'], account_name: str, email: str):
    """Add an account of type [ACCOUNT-TYPE] with the name [ACCOUNT-NAME]

    \b
    Example:
        $ rucio account add
    """
    ctx.obj.client.add_account(
        account=account_name,
        type_=account_type,
        email=email
    )
    print('Added new account: %s' % account)


@account.command("list")
@click.option("--type", "type_", type=click.Choice(["USER", "GROUP", "SERVICE"]))
@click.option("--id", help="Filter by identity (e.g. DN)")
@click.option("--filter", help="Filter arguments in form `key=value,another_key=next_value`")  # TODO Explicit numeration of these possible keys
@click.option('--csv', help="Output result as a CSV", is_flag=True)
@click.pass_context
def list_(ctx: click.Context, type_: Optional[Literal['USER', 'GROUP', 'SERVICE']], id: Optional[str], filter: Optional[str], csv: bool):
    """List all accounts that match given filters"""
    filters = {}
    if filter:
        for key, value in [(s.split('=')[0], s.split('=')[1]) for s in filter.split(',')]:
            filters[key] = value
    accounts = ctx.obj.client.list_accounts(identity=id, account_type=type_, filters=filters)
    if csv:
        print(*(account['account'] for account in accounts), sep=',')
    elif ctx.obj.use_rich:
        table = generate_table([
            [account['account']] for account in accounts],
            headers=['ACCOUNT'],
            col_alignments=['left']
        )
        ctx.obj.spinner.stop()
        print_output(table, console=ctx.obj.console, no_pager=ctx.obj.no_pager)
    else:
        for account in accounts:
            print(account['account'])


@account.command("show")
@click.argument("account-name")
@click.pass_context
def show(ctx: click.Context, account_name: str):
    """
    Show info about a single account
    """
    info = ctx.obj.client.get_account(account=account_name)
    if ctx.obj.use_rich:
        keyword_style = {**CLITheme.ACCOUNT_STATUS, **CLITheme.ACCOUNT_TYPE}
        table_data = [(k, Text(str(v), style=keyword_style.get(str(v), 'default'))) for k, v in sorted(info.items())]
        table = generate_table(
            table_data,
            row_styles=['none'],
            col_alignments=['left', 'left']
        )
        print_output(table, console=ctx.obj.console, no_pager=ctx.obj.no_pager)
    else:
        for k in info:
            print(k.ljust(10) + ' : ' + str(info[k]))


@account.command("remove")
@click.argument("account-name")
@click.pass_context
def remove(ctx: click.Context, account_name: str):
    """
    Remove an account
    (WARNING: Permanently disables the account. If you want to temporarily disable, use `account update [account-name] --ban`)
    """
    ctx.obj.client.delete_account(account_name)
    print('Deleted account: %s' % account_name)


@account.command("update")
@click.argument("account-name")
@click.option("--email", help="Email address associated with the account")
@click.option("--ban/--unban", default=None, help="Temporarily disable/enable an account")
@click.pass_context
def update(ctx: click.Context, ban: Optional[bool], account_name: str, email: Optional[str]):
    """Update account settings"""
    if ban is not None:
        if ban:
            ctx.obj.client.update_account(
                account=account_name, key='status', value='SUSPENDED'
            )
            print('Account %s banned' % account_name)
        else:
            ctx.obj.client.update_account(account=account_name, key='status', value='ACTIVE')
            print('Account %s unbanned' % account_name)
    else:
        ctx.obj.client.update_account(account=account_name, key='email', value=email)
        print('email of account %s changed to %s' % (account_name, email))


@account.group()
def attribute():
    """View or modify account attributes"""


@attribute.command("list")
@click.argument("account-name")
@click.pass_context
def attribute_list(ctx: click.Context, account_name: str):
    "List the attributes for a given account"
    attributes = next(ctx.obj.client.list_account_attributes(account_name))
    table_data = []
    for attr in attributes:
        table_data.append([attr['key'], attr['value']])

    if ctx.obj.use_rich:
        table = generate_table(table_data, headers=['Key', 'Value'], col_alignments=['left', 'left'])
        print_output(table, console=ctx.obj.console, no_pager=ctx.obj.no_pager)
    else:
        print(tabulate(table_data, tablefmt=ctx.obj.tablefmt, headers=['Key', 'Value']))


@attribute.command("add")
@click.argument("account-name")
@click.option('--key', help='Attribute key', required=True)
@click.option('--value', help='Attribute value', required=True)
@click.pass_context
def attribute_add(ctx: click.Context, account_name: str, key: str, value: str):
    """Add a new attribute [key] to an account"""
    ctx.obj.client.add_account_attribute(account=account_name, key=key, value=value)


@attribute.command("remove")
@click.argument("account-name")
@click.option("--key", help="Attribute key", required=True)
@click.pass_context
def attribute_remove(ctx: click.Context, account_name: str, key: str):
    """Remove an attribute from an account without reassigning it"""
    ctx.obj.client.delete_account_attribute(account=account_name, key=key)


@account.group("limit")
def limit():
    """View or modify account limits - limit how much data an account can store on an RSE"""


@limit.command("list")
@click.argument("account-name")
@click.option("--rse", "--rse-name", help="Show usage for only for this RSE.")
@click.pass_context
def limit_list(ctx: click.Context, account_name: str, rse: Optional[str]):
    """
    Shows the space used, the quota limit and the quota left
    for an account for every RSE where the user have quota.
    """
    if ctx.obj.use_rich:
        ctx.obj.spinner.update(status='Fetching account usage')
        ctx.obj.spinner.start()

    usage = ctx.obj.client.get_local_account_usage(account=account_name, rse=rse)
    table_data = []
    for item in usage:
        remaining = 0 if float(item['bytes_remaining']) < 0 else float(item['bytes_remaining'])
        table_data.append([item['rse'], sizefmt(item['bytes'], ctx.obj.human), sizefmt(item['bytes_limit'], ctx.obj.human), sizefmt(remaining, ctx.obj.human)])
    table_data.sort()

    if ctx.obj.use_rich:
        table1 = generate_table(table_data, headers=['RSE', 'USAGE', 'LIMIT', 'QUOTA LEFT'], col_alignments=['left', 'right', 'right', 'right'])
    else:
        print(tabulate(table_data, tablefmt=ctx.obj.tablefmt, headers=['RSE', 'USAGE', 'LIMIT', 'QUOTA LEFT']))

    table_data = []
    usage = ctx.obj.client.get_global_account_usage(account=account_name)
    for item in usage:
        if (rse and rse in item['rse_expression']) or not rse:
            remaining = 0 if float(item['bytes_remaining']) < 0 else float(item['bytes_remaining'])
            table_data.append([item['rse_expression'], sizefmt(item['bytes'], ctx.obj.human), sizefmt(item['bytes_limit'], ctx.obj.human), sizefmt(remaining, ctx.obj.human)])
    table_data.sort()

    if ctx.obj.use_rich:
        table2 = generate_table(table_data, headers=['RSE EXPRESSION', 'USAGE', 'LIMIT', 'QUOTA LEFT'], col_alignments=['left', 'right', 'right', 'right'])
    else:
        print(tabulate(table_data, tablefmt=ctx.obj.tablefmt, headers=['RSE EXPRESSION', 'USAGE', 'LIMIT', 'QUOTA LEFT']))

    if ctx.obj.use_rich:
        ctx.obj.spinner.stop()
        print_output(table1, table2, console=ctx.obj.console, no_pager=ctx.obj.no_pager)


@limit.command("add")
@click.argument(
    "account-name",
)
@click.option("--rse", "--rse-name", help="Full RSE name", required=True)  # TODO Separate RSE (local) and RSE Expression (global)
@click.option("--bytes", "bytes_", help='Value of the limit; can be specified in bytes ("10000"), with a storage unit ("10GB"), or "infinity"', required=True)
@click.option("--locality", type=click.Choice(["local", "global"]), help="Global or local limit scope", default="local")
@click.pass_context
def limit_add(ctx: click.Context, account_name: str, rse: str, bytes_: str, locality: str):
    """Add a new limit for an account on an RSE. An account can have both local and global limits on the same RSE."""
    byte_limit = None
    limit_input = bytes_.lower()

    if limit_input == 'inf' or limit_input == 'infinity':
        byte_limit = -1
    else:
        byte_limit = get_bytes_value_from_string(limit_input)
        if not byte_limit:
            try:
                byte_limit = int(limit_input)
            except ValueError:
                msg = f'\
                    The limit could not be set. Either you misspelled infinity or your input ({bytes_.lower()}) could not be converted to integer or you used a wrong pattern. \
                    Please use a format like 10GB with B,KB,MB,GB,TB,PB as units (not case sensitive)'
                raise InputValidationError(msg)

    ctx.obj.client.set_account_limit(account=account_name, rse=rse, bytes_=byte_limit, locality=locality.lower())
    print('Set account limit for account %s on RSE %s: %s' % (account_name, rse, sizefmt(byte_limit, True)))


@limit.command("remove")
@click.argument("account-name")
@click.option("--rse", "--rse-name", help="Full RSE name", required=True)
@click.option("--locality", type=click.Choice(["local", "global"]), help="Global or local limit scope", default="local")
@click.pass_context
def limit_remove(ctx: click.Context, account_name: str, rse: str, locality: str):
    """Remove existing limits for an account on an RSE"""
    ctx.obj.client.delete_account_limit(account=account_name, rse=rse, locality=locality)
    print('Deleted account limit for account %s and RSE %s' % (account_name, rse))


@account.group("identity")
def identity():
    """Manage identities for an account - used to login"""


@identity.command("list")
@click.argument("account-name", required=True)
@click.pass_context
def identity_list(ctx: click.Context, account_name: str):
    """See all the IDs for [account-name]"""
    table_data = []
    identities = ctx.obj.client.list_identities(account=account_name)
    for identity in identities:
        if ctx.obj.use_rich:
            table_data.append([identity['identity'], identity['type']])
        else:
            print('Identity: %(identity)s,\ttype: %(type)s' % identity)
    if ctx.obj.use_rich:
        table = generate_table(table_data, headers=['IDENTITY', 'TYPE'], col_alignments=['left', 'left'])
        print_output(table, console=ctx.obj.console, no_pager=ctx.obj.no_pager)


@identity.command("add")
@click.argument("account-name", required=True)
@click.option("--type", "type_", type=click.Choice(["X509", "GSS", "USERPASS", "SSH", "SAML", "OIDC"]), help="Authentication type", required=True)
@click.option("--id", help="Identity", required=True)
@click.option("--email", help="Email address associated with the identity", required=True)
@click.option("--password", help="Password if authtype is USERPASS")
@click.pass_context
def identity_add(ctx: click.Context, account_name: str, type_: str, id: str, email: str, password: Optional[str]):
    """Add a new identity for [account-name]"""
    if email == "":
        raise InputValidationError('Error: --email argument can\'t be an empty string. Failed to grant an identity access to an account')

    if type_ == 'USERPASS' and not password:
        raise InputValidationError('Missing --password argument')

    ctx.obj.client.add_identity(account=account_name, identity=id, authtype=type_, email=email, password=password)
    print('Added new identity to account: %s-%s' % (id, account_name))


@identity.command("remove")
@click.argument("account-name", required=True)
@click.option("--type", "type_", type=click.Choice(["X509", "GSS", "USERPASS", "SSH", "SAML", "OIDC"]), help="Authentication type", required=True)
@click.option("--id", help="Identity", required=True)
@click.pass_context
def identity_remove(ctx: click.Context, account_name: str, type_: str, id: str):
    """Revoke a given ID's access from an account"""
    ctx.obj.client.del_identity(account_name, id, authtype=type_)
    print('Deleted identity: %s' % id)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from logging import LogRecord
from logging.config import fileConfig
from pathlib import Path
from typing import TYPE_CHECKING, Any

from alembic import context
from sqlalchemy import engine_from_config, pool

if TYPE_CHECKING:
    from collections.abc import Collection, Mapping

    from alembic.runtime.migration import MigrationContext, MigrationInfo

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(config.config_file_name)

log = logging.getLogger("alembic.runtime.migration")


class _SuppressRunningLogFilter(logging.Filter):
    """Filter out Alembic's default "Running ..." log entries."""

    def filter(self, record: LogRecord) -> bool:
        message = record.getMessage()
        return not (message.startswith("Running upgrade ") or message.startswith("Running downgrade "))


log.addFilter(_SuppressRunningLogFilter())


def _log_version_details(
        ctx: 'MigrationContext',
        step: 'MigrationInfo',
        heads: 'Collection[Any]',
        run_args: 'Mapping[str, Any]',
) -> None:
    """Emit a precise log entry for each executed migration step."""

    revision = step.up_revision
    if revision is None:
        return

    operation = "upgrade" if step.is_upgrade else "downgrade"
    source = ",".join(step.source_revision_ids) or "base"
    destination = ",".join(step.destination_revision_ids) or "base"

    script_path = Path(revision.path).resolve()
    repo_root = Path(__file__).resolve().parent
    try:
        relative_path = script_path.relative_to(repo_root)
    except ValueError:
        relative_path = script_path

    log.info(f"Executing {operation} {source} -> {destination} using {relative_path}::{operation}")

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata


target_metadata = None

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:
    """
    Run migrations in 'offline' mode.

    This configures the context with just a URL/dialect
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.
    """

    # try getting url & version_table_schema
    url = config.get_main_option("sqlalchemy.url")
    version_table_schema = config.get_main_option("version_table_schema")

    # try getting dialect in case URL has not been specified
    dialect = config.get_main_option("dialect")

    # try getting starting revision
    starting_rev = config.get_main_option("starting_rev")

    context.configure(
        url=url,
        version_table_schema=version_table_schema,
        dialect_name=dialect,
        starting_rev=starting_rev,
        target_metadata=target_metadata,
        literal_binds=True,
        include_schemas=True,
        on_version_apply=_log_version_details,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """
    Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.
    """

    params = config.get_section(config.config_ini_section)

    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix='sqlalchemy.',
        poolclass=pool.NullPool)

    with connectable.connect() as connection:
        # Forcing the default is needed for PostgreSQL installations with named schemas.
        # For other databases it doesn't matter.
        # https://github.com/sqlalchemy/alembic/issues/409
        conn = connection.execution_options(schema_translate_map={None: params.get('version_table_schema', None)})

        context.configure(
            connection=conn,
            target_metadata=target_metadata,
            version_table_schema=params.get('version_table_schema', None),
            include_schemas=True,
            on_version_apply=_log_version_details)

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Remove RSE protocol default constraint"""    # noqa: D400, D415

from alembic import context
from alembic.op import alter_column

# Alembic revision identifiers
revision = '3b943000da18'
down_revision = '0548e199f091'

PRIORITY_COLUMNS = ['read_lan', 'write_lan', 'delete_lan',
                    'read_wan', 'write_wan', 'delete_wan',
                    'third_party_copy_read', 'third_party_copy_write']


def upgrade():
    """Upgrade the database to this revision."""
    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        schema = context.get_context().version_table_schema if context.get_context().version_table_schema else ''
        for column in PRIORITY_COLUMNS:
            alter_column('rse_protocols', column, server_default=None, schema=schema)


def downgrade():
    """Downgrade the database to the previous revision."""
    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        schema = context.get_context().version_table_schema if context.get_context().version_table_schema else ''
        for column in PRIORITY_COLUMNS:
            alter_column('rse_protocols', column, server_default='0', schema=schema)




'''
This file is automatically generated; Do not edit it. :)
'''
IS_FINAL = True
VERSION = '39.4.1'
BRANCH_NICK = 'release-39'
REVISION_ID = 'ec9c391d2816ac0b9fe0f45ed78bc8a05afcf8b2'
REVNO = '14337'



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys
from dataclasses import dataclass
from typing import Any, Optional
from urllib.parse import urlparse
from xml.etree import ElementTree  # noqa: S405 -- trusted XML input

import requests
from requests.adapters import HTTPAdapter
from urllib3.poolmanager import PoolManager

from rucio.common import exception
from rucio.common.constants import HTTPMethod
from rucio.rse.protocols import protocol


class TLSHTTPAdapter(HTTPAdapter):
    '''
    Class to force the SSL protocol to latest TLS
    '''
    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = PoolManager(num_pools=connections,
                                       maxsize=maxsize,
                                       block=block,
                                       cert_reqs="CERT_REQUIRED",
                                       ca_cert_dir="/etc/grid-security/certificates")


class UploadInChunks:
    '''
    Class to upload by chunks.
    '''

    def __init__(self, filename, chunksize, progressbar=False):
        self.__totalsize = os.path.getsize(filename)
        self.__readsofar = 0
        self.__filename = filename
        self.__chunksize = chunksize
        self.__progressbar = progressbar

    def __iter__(self):
        try:
            with open(self.__filename, 'rb') as file_in:
                while True:
                    data = file_in.read(self.__chunksize)
                    if not data:
                        if self.__progressbar:
                            sys.stdout.write("\n")
                        break
                    self.__readsofar += len(data)
                    if self.__progressbar:
                        percent = self.__readsofar * 100 / self.__totalsize
                        sys.stdout.write("\r{percent:3.0f}%".format(percent=percent))
                    yield data
        except OSError as error:
            raise exception.SourceNotFound(error)

    def __len__(self):
        return self.__totalsize


class IterableToFileAdapter:
    '''
    Class IterableToFileAdapter
    '''
    def __init__(self, iterable):
        self.iterator = iter(iterable)
        self.length = len(iterable)

    def read(self, size=-1):   # TBD: add buffer for `len(data) > size` case
        nextvar = next(self.iterator, b'')
        return nextvar

    def __len__(self):
        return self.length


@dataclass(frozen=True)
class _PropfindFile:
    """Contains the properties of one file from a PROPFIND response."""

    href: str
    size: Optional[int]

    @classmethod
    def from_xml_node(cls, node: ElementTree.Element):
        """Extract file properties from a `<{DAV:}response>` node."""

        xml_href = node.find('./{DAV:}href')
        if xml_href is None or xml_href.text is None:
            raise ValueError('Response is missing mandatory field "href".')
        else:
            href = xml_href.text

        xml_size = node.find('./{DAV:}propstat/{DAV:}prop/{DAV:}getcontentlength')
        if xml_size is None or xml_size.text is None:
            size = None
        else:
            size = int(xml_size.text)

        return cls(href=href, size=size)  # type: ignore


@dataclass(frozen=True)
class _PropfindResponse:
    """Contains all the files from a PROPFIND response."""

    files: tuple[_PropfindFile]

    @classmethod
    def parse(cls, document: str):
        """Parses the XML document of a WebDAV PROPFIND response.

        The PROPFIND response is described in RFC 4918.
        This method expects the document root to be a node with tag `{DAV:}multistatus`.

        :param document: XML document to parse.
        :raises ValueError: if the XML document couldn't be parsed.
        :returns: The parsed response.
        """

        try:
            xml = ElementTree.fromstring(document)  # noqa: S314
        except ElementTree.ParseError as ex:
            raise ValueError("Couldn't parse XML document") from ex

        if xml.tag != '{DAV:}multistatus':
            raise ValueError('Root element is not "{DAV:}multistatus".')

        files = []
        for xml_response in xml.findall('./{DAV:}response'):
            files.append(_PropfindFile.from_xml_node(xml_response))

        return cls(files=tuple(files))  # type: ignore


class Default(protocol.RSEProtocol):

    """ Implementing access to RSEs using the webDAV protocol."""

    def connect(self, credentials: Optional[dict[str, Any]] = None) -> None:
        """ Establishes the actual connection to the referred RSE.

            :param credentials: Provides information to establish a connection
                to the referred storage system. For WebDAV connections these are
                ca_cert, cert, auth_type, timeout

            :raises RSEAccessDenied
        """
        credentials = credentials or {}
        using_presigned_urls = self.rse['sign_url'] is not None
        try:
            parse_url = urlparse(self.path2pfn(''))
            self.server = f'{parse_url.scheme}://{parse_url.netloc}'
        except KeyError:
            raise exception.RSEAccessDenied('No specified Server')

        try:
            self.ca_cert = credentials['ca_cert']
        except KeyError:
            self.ca_cert = None

        try:
            self.auth_type = credentials['auth_type']
        except KeyError:
            self.auth_type = 'cert'

        if using_presigned_urls:
            # Suppress all authentication, otherwise S3 servers will reject
            # requests.
            self.cert = None
            self.auth_token = None
        else:
            try:
                self.cert = credentials['cert']
            except KeyError:
                x509 = os.getenv('X509_USER_PROXY')
                if not x509:
                    # Trying to get the proxy from the default location
                    proxy_path = '/tmp/x509up_u%s' % os.geteuid()
                    if os.path.isfile(proxy_path):
                        self.cert = (proxy_path, proxy_path)
                    elif self.auth_token:
                        # If no proxy is found, we set the cert to None and use the auth_token
                        self.cert = None
                        pass
                    else:
                        raise exception.RSEAccessDenied('X509_USER_PROXY is not set')
                else:
                    self.cert = (x509, x509)

        try:
            self.timeout = credentials['timeout']
        except KeyError:
            self.timeout = 300
        self.session = requests.Session()
        self.session.mount('https://', TLSHTTPAdapter())
        if self.auth_token:
            self.session.headers.update({'Authorization': 'Bearer ' + self.auth_token})
        # "ping" to see if the server is available
        try:
            test_url = self.path2pfn('')
            res = self.session.request('HEAD', test_url, verify=False, timeout=self.timeout, cert=self.cert)
            # REVISIT: this test checks some URL that doesn't correspond to
            # any valid Rucio file.  Although this works for normal WebDAV
            # endpoints, it fails for endpoints using presigned URLs.  As a
            # work-around, accept 4xx status codes when using presigned URLs.
            if res.status_code != 200 and not (using_presigned_urls and res.status_code < 500):
                raise exception.ServiceUnavailable('Bad status code %s %s : %s' % (res.status_code, test_url, res.text))
        except requests.exceptions.ConnectionError as error:
            raise exception.ServiceUnavailable('Problem to connect %s : %s' % (test_url, error))
        except requests.exceptions.ReadTimeout as error:
            raise exception.ServiceUnavailable(error)

    def close(self):
        self.session.close()

    def path2pfn(self, path):
        """
            Returns a fully qualified PFN for the file referred by path.

            :param path: The path to the file.

            :returns: Fully qualified PFN.

        """
        if not path.startswith('https'):
            return '%s://%s:%s%s%s' % (self.attributes['scheme'], self.attributes['hostname'], str(self.attributes['port']), self.attributes['prefix'], path)
        else:
            return path

    def exists(self, pfn):
        """ Checks if the requested file is known by the referred RSE.

            :param pfn: Physical file name

            :returns: True if the file exists, False if it doesn't

            :raise  ServiceUnavailable, RSEAccessDenied
        """
        path = self.path2pfn(pfn)

        using_presigned_urls = self.rse['sign_url'] is not None

        try:
            # use GET instead of HEAD for presigned urls
            if not using_presigned_urls:
                result = self.session.request(HTTPMethod.HEAD.value, path, verify=False, timeout=self.timeout,
                                              cert=self.cert)
            else:
                result = self.session.request(HTTPMethod.GET.value, path, verify=False, timeout=self.timeout,
                                              cert=self.cert)
            if result.status_code == 200:
                return True
            elif result.status_code in [401, ]:
                raise exception.RSEAccessDenied()
            elif result.status_code in [404, ]:
                return False
            else:
                # catchall exception
                raise exception.RucioException(result.status_code, result.text)
        except requests.exceptions.ConnectionError as error:
            raise exception.ServiceUnavailable(error)

    def get(self, pfn, dest='.', transfer_timeout=None):
        """ Provides access to files stored inside connected the RSE.

            :param pfn: Physical file name of requested file
            :param dest: Name and path of the files when stored at the client
            :param transfer_timeout: Transfer timeout (in seconds)

            :raises DestinationNotAccessible, ServiceUnavailable, SourceNotFound, RSEAccessDenied
        """
        path = self.path2pfn(pfn)
        chunksize = 1024
        transfer_timeout = self.timeout if transfer_timeout is None else transfer_timeout

        try:
            result = self.session.get(path, verify=False, stream=True, timeout=transfer_timeout, cert=self.cert)
            if result and result.status_code in [200, ]:
                length = None
                if 'content-length' in result.headers:
                    length = int(result.headers['content-length'])
                with open(dest, 'wb') as file_out:
                    nchunk = 0
                    if not length:
                        print('Malformed HTTP response (missing content-length header).')
                    for chunk in result.iter_content(chunksize):
                        file_out.write(chunk)
                        if length:
                            nchunk += 1
            elif result.status_code in [404, ]:
                raise exception.SourceNotFound()
            elif result.status_code in [401, 403]:
                raise exception.RSEAccessDenied()
            else:
                # catchall exception
                raise exception.RucioException(result.status_code, result.text)
        except requests.exceptions.ConnectionError as error:
            raise exception.ServiceUnavailable(error)
        except requests.exceptions.ReadTimeout as error:
            raise exception.ServiceUnavailable(error)

    def put(self, source, target, source_dir=None, transfer_timeout=None, progressbar=False):
        """ Allows to store files inside the referred RSE.

            :param source: Physical file name
            :param target: Name of the file on the storage system e.g. with prefixed scope
            :param source_dir Path where the to be transferred files are stored in the local file system
            :param transfer_timeout Transfer timeout (in seconds)

            :raises DestinationNotAccessible, ServiceUnavailable, SourceNotFound, RSEAccessDenied
        """
        path = self.path2pfn(target)
        full_name = source_dir + '/' + source if source_dir else source
        directories = path.split('/')
        # Try the upload without testing the existence of the destination directory
        transfer_timeout = self.timeout if transfer_timeout is None else transfer_timeout

        try:
            if not os.path.exists(full_name):
                raise exception.SourceNotFound()
            it = UploadInChunks(full_name, 10000000, progressbar)
            result = self.session.put(path, data=IterableToFileAdapter(it), verify=False, allow_redirects=True, timeout=transfer_timeout, cert=self.cert)
            if result.status_code in [200, 201]:
                return
            if result.status_code in [409, ]:
                raise exception.FileReplicaAlreadyExists()
            else:
                # Create the directories before issuing the PUT
                for directory_level in reversed(list(range(1, 4))):
                    upper_directory = "/".join(directories[:-directory_level])
                    self.mkdir(upper_directory)
                try:
                    if not os.path.exists(full_name):
                        raise exception.SourceNotFound()
                    it = UploadInChunks(full_name, 10000000, progressbar)
                    result = self.session.put(path, data=IterableToFileAdapter(it), verify=False, allow_redirects=True, timeout=transfer_timeout, cert=self.cert)
                    if result.status_code in [200, 201]:
                        return
                    if result.status_code in [409, ]:
                        raise exception.FileReplicaAlreadyExists()
                    elif result.status_code in [401, ]:
                        raise exception.RSEAccessDenied()
                    else:
                        # catchall exception
                        raise exception.RucioException(result.status_code, result.text)
                except requests.exceptions.ConnectionError as error:
                    raise exception.ServiceUnavailable(error)
                except OSError as error:
                    raise exception.SourceNotFound(error)
        except requests.exceptions.ConnectionError as error:
            raise exception.ServiceUnavailable(error)
        except requests.exceptions.ReadTimeout as error:
            raise exception.ServiceUnavailable(error)
        except OSError as error:
            raise exception.SourceNotFound(error)

    def rename(self, pfn, new_pfn):
        """ Allows to rename a file stored inside the connected RSE.

            :param pfn:      Current physical file name
            :param new_pfn  New physical file name

            :raises DestinationNotAccessible, ServiceUnavailable, SourceNotFound, RSEAccessDenied
        """
        path = self.path2pfn(pfn)
        new_path = self.path2pfn(new_pfn)
        directories = new_path.split('/')

        headers = {'Destination': new_path}
        # Try the rename without testing the existence of the destination directory
        try:
            result = self.session.request('MOVE', path, verify=False, headers=headers, timeout=self.timeout, cert=self.cert)
            if result.status_code == 201:
                return
            elif result.status_code in [404, ]:
                raise exception.SourceNotFound()
            else:
                # Create the directories before issuing the MOVE
                for directory_level in reversed(list(range(1, 4))):
                    upper_directory = "/".join(directories[:-directory_level])
                    self.mkdir(upper_directory)
                try:
                    result = self.session.request('MOVE', path, verify=False, headers=headers, timeout=self.timeout, cert=self.cert)
                    if result.status_code == 201:
                        return
                    elif result.status_code in [404, ]:
                        raise exception.SourceNotFound()
                    elif result.status_code in [401, ]:
                        raise exception.RSEAccessDenied()
                    else:
                        # catchall exception
                        raise exception.RucioException(result.status_code, result.text)
                except requests.exceptions.ConnectionError as error:
                    raise exception.ServiceUnavailable(error)
        except requests.exceptions.ConnectionError as error:
            raise exception.ServiceUnavailable(error)
        except requests.exceptions.ReadTimeout as error:
            raise exception.ServiceUnavailable(error)

    def delete(self, pfn):
        """ Deletes a file from the connected RSE.

            :param pfn: Physical file name

            :raises ServiceUnavailable, SourceNotFound, RSEAccessDenied, ResourceTemporaryUnavailable
        """
        path = self.path2pfn(pfn)
        try:
            result = self.session.delete(path, verify=False, timeout=self.timeout, cert=self.cert)
            if result.status_code in [204, ]:
                return
            elif result.status_code in [404, ]:
                raise exception.SourceNotFound()
            elif result.status_code in [401, 403]:
                raise exception.RSEAccessDenied()
            elif result.status_code in [500, 503]:
                raise exception.ResourceTemporaryUnavailable()
            else:
                # catchall exception
                raise exception.RucioException(result.status_code, result.text)
        except requests.exceptions.ConnectionError as error:
            raise exception.ServiceUnavailable(error)
        except requests.exceptions.ReadTimeout as error:
            raise exception.ServiceUnavailable(error)

    def mkdir(self, directory):
        """ Internal method to create directories

            :param directory: Name of the directory that needs to be created

            :raises DestinationNotAccessible, ServiceUnavailable, SourceNotFound, RSEAccessDenied
        """
        path = self.path2pfn(directory)
        try:
            result = self.session.request('MKCOL', path, verify=False, timeout=self.timeout, cert=self.cert)
            if result.status_code in [201, 405]:  # Success or directory already exists
                return
            elif result.status_code in [404, ]:
                raise exception.SourceNotFound()
            elif result.status_code in [401, ]:
                raise exception.RSEAccessDenied()
            else:
                # catchall exception
                raise exception.RucioException(result.status_code, result.text)
        except requests.exceptions.ConnectionError as error:
            raise exception.ServiceUnavailable(error)
        except requests.exceptions.ReadTimeout as error:
            raise exception.ServiceUnavailable(error)

    def ls(self, filename):
        """ Internal method to list files/directories

            :param filename: Name of the directory that needs to be created

            :raises DestinationNotAccessible, ServiceUnavailable, SourceNotFound, RSEAccessDenied
        """
        path = self.path2pfn(filename)
        headers = {'Depth': '1'}
        self.exists(filename)
        try:
            result = self.session.request('PROPFIND', path, verify=False, headers=headers, timeout=self.timeout, cert=self.cert)
            if result.status_code in [404, ]:
                raise exception.SourceNotFound()
            elif result.status_code in [401, ]:
                raise exception.RSEAccessDenied()

            try:
                propfind = _PropfindResponse.parse(result.text)
            except ValueError:
                raise exception.ServiceUnavailable("Couldn't parse WebDAV response.")

            list_files = [self.server + file.href for file in propfind.files if file.href is not None]

            try:
                list_files.remove(filename + '/')
            except ValueError:
                pass
            try:
                list_files.remove(filename)
            except ValueError:
                pass

            return list_files
        except requests.exceptions.ConnectionError as error:
            raise exception.ServiceUnavailable(error)
        except requests.exceptions.ReadTimeout as error:
            raise exception.ServiceUnavailable(error)

    def stat(self, path):
        """
            Returns the stats of a file.

            :param path: path to file

            :raises ServiceUnavailable: if some generic error occurred in the library.
            :raises SourceNotFound: if the source file was not found on the referred storage.
            :raises RSEAccessDenied: in case of permission issue.

            :returns: a dict with filesize of the file provided in path as a key.
        """
        headers = {'Depth': '1'}
        dict_ = {}
        try:
            result = self.session.request('PROPFIND', path, verify=False, headers=headers, timeout=self.timeout, cert=self.cert)
            if result.status_code in [404, ]:
                raise exception.SourceNotFound()
            elif result.status_code in [401, ]:
                raise exception.RSEAccessDenied()
            if result.status_code in [400, ]:
                raise exception.InvalidRequest()
        except requests.exceptions.ConnectionError as error:
            raise exception.ServiceUnavailable(error)
        except requests.exceptions.ReadTimeout as error:
            raise exception.ServiceUnavailable(error)

        path_parts = self.parse_pfns(path)[path]
        local_path = os.path.join(path_parts['prefix'], path_parts['path'][1:], path_parts['name'])

        try:
            propfind = _PropfindResponse.parse(result.text)
        except ValueError:
            raise exception.ServiceUnavailable("Couldn't parse WebDAV response.")

        for file in propfind.files:
            if file.href != str(local_path):
                continue

            if file.size is None:
                continue

            dict_['filesize'] = file.size
            break
        else:
            raise exception.ServiceUnavailable("WebDAV response didn't include content length for requested path.")

        return dict_

    def get_space_usage(self):
        """
        Get RSE space usage information.

        :returns: a list with dict containing 'totalsize' and 'unusedsize'

        :raises ServiceUnavailable: if some generic error occurred in the library.
        """
        endpoint_basepath = self.path2pfn('')
        headers = {'Depth': '0'}

        try:
            root = ElementTree.fromstring(self.session.request('PROPFIND', endpoint_basepath, verify=False, headers=headers, cert=self.session.cert).text)  # noqa: S314
            usedsize = root[0][1][0].find('{DAV:}quota-used-bytes').text
            try:
                unusedsize = root[0][1][0].find('{DAV:}quota-available-bytes').text
            except Exception:
                print('No free space given, return -999')
                unusedsize = -999
            totalsize = int(usedsize) + int(unusedsize)
            return totalsize, unusedsize
        except Exception as error:
            raise exception.ServiceUnavailable(error)


class NoRename(Default):
    """ Implementing access to RSEs using the WebDAV protocol but without
    renaming files on upload/download. Necessary for some storage endpoints.
    """

    def __init__(self, protocol_attr, rse_settings, logger=logging.log):
        """ Initializes the object with information about the referred RSE.

            :param protocol_attr:  Properties of the requested protocol.
            :param rse_settings:   The RSE settings.
            :param logger:         Optional decorated logger that can be passed from the calling daemons or servers.
        """
        super(NoRename, self).__init__(protocol_attr, rse_settings, logger=logger)
        self.renaming = False
        self.attributes.pop('determinism_type', None)

    def rename(self, pfn, new_pfn):
        """ Allows to rename a file stored inside the connected RSE.

            :param pfn:      Current physical file name
            :param new_pfn  New physical file name

            :raises DestinationNotAccessible, ServiceUnavailable, SourceNotFound
        """
        raise NotImplementedError



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Opendata Record ID"""  # noqa: D400, D415

import sqlalchemy as sa
from alembic import op

from rucio.common.schema import get_schema_value

# Alembic revision identifiers
revision = 'a7e76cf4881d'
down_revision = 'a62db546a1f1'


def upgrade():
    op.create_table(
        'dids_opendata_record',
        sa.Column('scope', sa.String(length=get_schema_value('SCOPE_LENGTH')), nullable=False),
        sa.Column('name', sa.String(length=get_schema_value('NAME_LENGTH')), nullable=False),
        sa.Column('record_id', sa.Integer(), nullable=False, unique=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('scope', 'name', name='OPENDATA_RECORD_PK'),
        sa.ForeignKeyConstraint(['scope', 'name'], ['dids_opendata.scope', 'dids_opendata.name'],
                                ondelete='CASCADE', name='OPENDATA_RECORD_FK')
    )
    op.create_index('OPENDATA_RECORD_CREATED_AT_IDX', 'dids_opendata_record', ['created_at'])
    op.create_index('OPENDATA_RECORD_UPDATED_AT_IDX', 'dids_opendata_record', ['updated_at'])


def downgrade():
    op.drop_index('OPENDATA_RECORD_CREATED_AT_IDX', table_name='dids_opendata_record')
    op.drop_index('OPENDATA_RECORD_UPDATED_AT_IDX', table_name='dids_opendata_record')
    op.drop_table('dids_opendata_record')
